#include <ArduinoLowPower.h>
#define PIN_SBU_RX 7
#define PIN_SBU_TX 6
#define PIN_LED 13
#define INT_PIN 5

// LED is inverted
#define OFF HIGH
#define ON LOW

uint32_t active = 0;

void setup(){
	pinMode(PIN_LED,OUTPUT);
	digitalWrite(PIN_LED, OFF);
	Serial.begin(115200);
	pinMode(PIN_SBU_RX,INPUT);
	active = millis();
	LowPower.attachInterruptWakeup(INT_PIN, wakeup, CHANGE);

}

void wakeup(){
	active = millis();
}

void loop(){
if(digitalRead(PIN_SBU_RX) == HIGH){
	
	Serial1.begin(19200);
	do { uint32_t i = process(); 
		if(i){
			digitalWrite(PIN_LED, ON);
			active=i;
			}
	}
	while (millis()-active < 10000); // stay on for 10 seconds
	Serial1.end();
}
digitalWrite(PIN_LED, OFF);
if(millis()-active > 10000){
	LowPower.deepSleep();
	
	//fakesleep(true);
	//while(fakesleep(false)){};
}
}

// Test Sleepmode without losing USB serial connection
bool fakesleep(bool init){
	static bool stat = 0;
	if(init){ 
		stat = digitalRead(PIN_SBU_RX); 
		Serial.println("OFF");
	}
	if( digitalRead(PIN_SBU_RX) != stat){
		stat = !stat;
		active = millis();
		Serial.println("ON");
		return false;
	}
	return true;
}


uint32_t process(){
	static uint8_t bl_status = 0; // status of the red backlight (just emulated)
	static uint8_t ch1 [2];
	static bool ch1_set = false; // first challenge received and stored
	static int rx_len = 0;
	uint8_t rx_d[250];
	static uint32_t last_d = 0; // last millis() serial data was received
	static int SN; // sequence number
	static int CMD; // command
	static int PID;
	uint8_t tx_d[8];
	uint8_t tx_len =0;
	bool proc = false;

	uint32_t diff = millis() - last_d;
	if (diff > 5 && rx_len){ // get one message, wait for >5ms spacing
		proc = true;
		last_d=0;
		print_data(rx_d, rx_len, true);
	}

	if(Serial1.available()){
		rx_d[rx_len] = Serial1.read();
		rx_len ++;
		last_d = millis();
	}

	if(proc){
	// Process data 
		uint8_t c = ccrc(rx_d ,rx_len-1);
		if( c == rx_d[rx_len-1]){ // check CRC

			if(rx_d[0] == 0x40){ ch1_set = false;} // rest on first init message

			SN =  rx_d[0] & 0x0F;
			CMD = rx_d[0] & 0xF0;
			PID = rx_d[1];

			tx_d[0] = (SN ^ CMD) + 0X20;
			tx_d[1] = PID;

			if(CMD == 0x40) {
				if(PID == 0x01) {
					if(!ch1_set){ // first stage
						ch1[0] = rx_d[2];
						ch1[1] = rx_d[3];
						ch1_set = true;
						// send own challenge
						tx_d[2] = 0x55;
						tx_d[3] = 0xEC;
						tx_len = 4;
						send_data(tx_d,tx_len);

					} else { // second stage
						// check challenge here, but I don't care

						// Prepare response
						tx_d[2] = count_ones(ch1);
						tx_d[3] = ccrc(ch1, 2);
						tx_len = 4;
						send_data(tx_d,tx_len);
						ch1_set = false;
					}
				}
				if(PID == 0x05) {
					tx_d[2] = 0x32;
					tx_d[3] = 0x2A;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x0D) { // set Backlight default
					tx_d[2] = 0x00;
					tx_d[3] = bl_status;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x0E) { // set Backlight
					bl_status = rx_d[3];
					tx_d[2] = 0x00;
					tx_d[3] = bl_status;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
			}
			else if(CMD == 0x00) {
				if(PID == 0x04) {
					tx_d[2] = 0x0B;
					tx_d[3] = 0x86;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x12) {
					tx_d[2] = 0x0B;
					tx_d[3] = 0x86;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x0C) {
					tx_d[2] = 0x00;
					tx_d[3] = 0x00;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x11) {
					tx_d[2] = 0x01;
					tx_d[3] = 0x10;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x82) {
					tx_d[2] = 0xEC;
					tx_d[3] = 0xF6;
					tx_d[4] = 0x00;
					tx_d[5] = 0x0A;
					tx_d[6] = 0x12;
					tx_d[7] = 0x3C;
					tx_d[8] = 0x7F;
					tx_d[9] = 0x7F;
					tx_len = 10;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x83) {
					tx_d[2] = 0x0C;
					tx_d[3] = 0x10;
					tx_d[4] = 0x18;
					tx_d[5] = 0x22;
					tx_d[6] = 0x32;
					tx_d[7] = 0x32;
					tx_d[8] = 0xFF;
					tx_d[9] = 0xFF;
					tx_len = 10;
					send_data(tx_d,tx_len);

				}
				if(PID == 0x07) {
					tx_d[2] = 0x0A;
					tx_d[3] = 0xD2;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x84) {
					tx_d[2] = 0x00;
					tx_d[3] = 0x1E;
					tx_d[4] = 0x00;
					tx_d[5] = 0x64;
					tx_d[6] = 0x00;
					tx_d[7] = 0x00;
					tx_d[8] = 0x00;
					tx_d[9] = 0x00;
					tx_len = 10;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x0A) { // Voltage
					tx_d[2] = 0x8B;
					tx_d[3] = 0xC8;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x80) { // Stat (currently static content)
					tx_d[2] = 0x0F;
					tx_d[3] = 0x3E;
					tx_d[4] = 0xFF;
					tx_d[5] = 0xA6;
					tx_d[6] = 0x8B;
					tx_d[7] = 0x18;
					tx_d[8] = 0x01;
					tx_d[9] = 0x06;
					tx_len = 10;
					send_data(tx_d,tx_len);
				}
				if(PID == 0x0D) { // Backlight
					tx_d[2] = 0x00;
					tx_d[3] = bl_status;
					tx_len = 4;
					send_data(tx_d,tx_len);
				}
			}
		}
	proc = false;
	rx_len=0;
	}
	return last_d;
}

void send_data(uint8_t *data, int len){
	static uint32_t last_send = 0;
	while(millis()-last_send < 5){delay(1);} // generate TX spacing if required
	uint8_t c = ccrc(data, len);
	Serial1.write(data, len);
	Serial1.write(&c, 1); // write crc
	Serial.print("Send: ");
	print_data(data, len, false);
	Serial.printf("%02X\n" ,c);
	last_send = millis();
}

uint8_t count_ones(uint8_t *data){ //16bit
	int count = 0;
	uint16_t n = data[1] | (data[0] << 8);
    for (int i = 0; i < 16; i++)
    {
        count += (n & 0x01);    // check last bit
        n >>= 1;
    }
    return count;
}


uint8_t ccrc (uint8_t *data, int len){
    uint8_t crc = 0xff;
    size_t i, j;
    for (i = 0; i < len; i++) {
        crc ^= data[i];
        //Serial.println(crc);
        for (j = 0; j < 8; j++) {
            if ((crc & 0x80) != 0)
                crc = (uint8_t)((crc << 1) ^ 0x31);
            else
                crc <<= 1;
        }
    }
    return crc;
}


void print_data(uint8_t * data, int count, bool nl){
	if(count){
		for (int i = 0; i<count; i++ ) {
			Serial.printf("%02X " ,data[i]);
		}
		if(nl){Serial.println();}
	}
}